import asyncio
import binascii
import functools
import hashlib
import os
import pathlib
import shutil
import subprocess
from datetime import datetime, timezone

import pytest

import py7zr

try:
    import libarchive.public as LAPublic
except ImportError:
    LAPublic = None

testdata_path = os.path.join(os.path.dirname(__file__), "data")
os.umask(0o022)


def check_output(expected, tmpdir):
    for exp in expected:
        if isinstance(tmpdir, str):
            target = pathlib.Path(tmpdir).joinpath(exp["filename"])
        else:
            target = tmpdir.joinpath(exp["filename"])
        if os.name == "posix":
            if exp.get("mode", None):
                assert target.stat().st_mode == exp["mode"], "%s, actual: %d, expected: %d" % (
                    exp["filename"],
                    target.stat().st_mode,
                    exp["mode"],
                )
        if exp.get("mtime", None):
            assert target.stat().st_mtime == exp["mtime"], "%s, actual: %d, expected: %d" % (
                exp["filename"],
                target.stat().st_mtime,
                exp["mtime"],
            )
        m = hashlib.sha256()
        m.update(target.open("rb").read())
        assert m.digest() == binascii.unhexlify(exp["digest"]), "Fails digest for %s" % exp["filename"]


def decode_all(archive, expected, tmpdir):
    for i, file_info in enumerate(archive.files):
        assert file_info.lastwritetime is not None
        assert file_info.filename is not None
    archive.extractall(path=tmpdir)
    archive.close()
    check_output(expected, tmpdir)


async def aio7zr(archive, path):
    loop = asyncio.get_event_loop()
    sevenzip = py7zr.SevenZipFile(archive)
    partial_py7zr = functools.partial(sevenzip.extractall, path=path)
    loop.run_in_executor(None, partial_py7zr)
    loop.run_in_executor(None, sevenzip.close)


def ltime(dt_utc):
    return dt_utc.replace(tzinfo=timezone.utc).astimezone(tz=None).strftime("%Y-%m-%d %H:%M:%S")


def ltime2(y, m, d, h, min, s):
    return ltime(datetime(y, m, d, h, min, s))


def p7zip_test(archive):
    if shutil.which("7z"):
        result = subprocess.run(["7z", "t", archive.as_posix()], stdout=subprocess.PIPE)
        if result.returncode != 0:
            print(result.stdout)
            pytest.fail("7z command report error")


def libarchive_extract(archive, tmpdir):
    if LAPublic:
        if not tmpdir.exists():
            tmpdir.mkdir(parents=True)
        with LAPublic.file_reader(str(archive)) as e:
            for entry in e:
                if entry.filetype.IFDIR:
                    tmpdir.joinpath(entry.pathname).mkdir(parents=True)
                elif entry.filetype.IFLNK:
                    tmpdir.joinpath(entry.pathname).link_to(entry.symlink_targetpath)
                else:
                    with tmpdir.joinpath(entry.pathname).open(mode="wb") as f:
                        for block in entry.get_blocks():
                            f.write(block)



import os
import pathlib
from typing import Optional

import requests
from pytest_httpserver import HTTPServer

import py7zr
import py7zr.io

testdata_path = pathlib.Path(os.path.dirname(__file__)).joinpath("data")
os.umask(0o022)


def test_extract_stream(httpserver: HTTPServer):
    httpserver.expect_request("/setup.cfg", method="PUT").respond_with_data("ok")
    httpserver.expect_request("/setup.cfg", method="GET").respond_with_data(
        b"[flake8]\nmax-line-length = 125\n\n[bdist_wheel]\nuniversal=1\n"
    )
    httpserver.expect_request("/setup.py", method="PUT").respond_with_data("ok")
    httpserver.expect_request("/scripts/py7zr", method="PUT").respond_with_data("ok")
    factory = StreamWriterFactory(httpserver)
    with py7zr.SevenZipFile(testdata_path.joinpath("test_1.7z").open(mode="rb")) as archive:
        archive.extractall(factory=factory)
    assert len(factory.products) == 3
    assert "setup.cfg" in factory.products.keys()
    assert "setup.py" in factory.products.keys()
    assert "scripts/py7zr" in factory.products.keys()
    assert factory.products["setup.cfg"].read() == b"[flake8]\nmax-line-length = 125\n\n[bdist_wheel]\nuniversal=1\n"


class StreamWriter(py7zr.io.Py7zIO):
    """Pseudo object storage writer."""

    def __init__(self, httpserver: HTTPServer, fname: str):
        self.httpserver: HTTPServer = httpserver
        self.fname = fname
        self.length = 0

    def write(self, data: [bytes, bytearray]):
        self.length += len(data)
        requests.put(self.httpserver.url_for(self.fname), data=data)
        self.httpserver.check_assertions()

    def read(self, size: Optional[int] = None) -> bytes:
        response = requests.get(self.httpserver.url_for(self.fname))
        return response.content

    def seek(self, offset: int, whence: int = 0) -> int:
        return offset

    def flush(self) -> None:
        pass

    def size(self) -> int:
        return self.length


class StreamWriterFactory(py7zr.io.WriterFactory):
    """Factory class to return StreamWriter object."""

    def __init__(self, httpserver: HTTPServer):
        self.httpserver: HTTPServer = httpserver
        self.products = {}

    def create(self, filename: str) -> py7zr.io.Py7zIO:
        product = StreamWriter(self.httpserver, filename)
        self.products[filename] = product
        return product



import binascii
import ctypes
import filecmp
import hashlib
import io
import lzma
import os
import pathlib
import shutil
import stat
import sys
from datetime import datetime, timezone

import pytest

import py7zr.archiveinfo
import py7zr.compressor
import py7zr.helpers
import py7zr.properties
from py7zr import SevenZipFile, pack_7zarchive
from py7zr.py7zr import FILE_ATTRIBUTE_UNIX_EXTENSION

from . import libarchive_extract, ltime, p7zip_test

testdata_path = os.path.join(os.path.dirname(__file__), "data")


def check_bit(val, mask):
    return val & mask == mask


@pytest.mark.basic
def test_compress_single_encoded_header(capsys, tmp_path):
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(True)
    archive.writeall(os.path.join(testdata_path, "test1.txt"), "test1.txt")
    assert len(archive.files) == 1
    archive.close()
    with target.open("rb") as target_archive:
        val = target_archive.read(1000)
        assert val.startswith(py7zr.properties.MAGIC_7Z)
    archive = py7zr.SevenZipFile(target, "r")
    assert archive.testzip() is None
    archive.close()
    mtime = datetime.fromtimestamp(pathlib.Path(os.path.join(testdata_path, "test1.txt")).stat().st_mtime, timezone.utc)
    expected = (
        "total 1 files and directories in archive\n"
        "   Date      Time    Attr         Size   Compressed  Name\n"
        "------------------- ----- ------------ ------------  ------------------------\n"
    )
    expected += "{} ....A           33           37  test1.txt\n".format(ltime(mtime))
    expected += "------------------- ----- ------------ ------------  ------------------------\n"
    cli = py7zr.cli.Cli()
    cli.run(["l", str(target)])
    out, err = capsys.readouterr()
    assert expected == out
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_directory_encoded_header(tmp_path):
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(True)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    assert len(archive.files) == 2
    archive._write_flush()
    assert archive.header.main_streams.packinfo.numstreams == 1
    assert archive.header.main_streams.packinfo.packsizes == [17]
    assert archive.header.main_streams.unpackinfo.numfolders == 1
    assert len(archive.header.main_streams.unpackinfo.folders) == 1
    assert len(archive.header.main_streams.unpackinfo.folders[0].coders) == 2
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.substreamsinfo.unpacksizes == [11]
    assert len(archive.header.files_info.files) == 2
    archive._fpclose()
    with target.open("rb") as target_archive:
        val = target_archive.read(1000)
        assert val.startswith(py7zr.properties.MAGIC_7Z)
    archive = py7zr.SevenZipFile(target, "r")
    assert archive.testzip() is None
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_files_encoded_header(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_1.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(True)
    archive.writeall(".")
    archive._write_flush()
    assert len(archive.files) == 4
    assert len(archive.header.files_info.files) == 4
    expected = [True, False, False, False]
    for i, f in enumerate(archive.header.files_info.files):
        f["emptystream"] = expected[i]
    assert archive.header.files_info.emptyfiles == [True, False, False, False]
    assert archive.header.files_info.files[3]["emptystream"] is False
    expected_attributes = stat.FILE_ATTRIBUTE_ARCHIVE
    if os.name == "posix":
        expected_attributes |= 0x8000 | (0o644 << 16)
    assert archive.header.files_info.files[3]["attributes"] == expected_attributes
    assert archive.header.files_info.files[3]["maxsize"] == 441
    assert archive.header.files_info.files[3]["uncompressed"] == 559
    assert archive.header.main_streams.packinfo.numstreams == 1
    assert archive.header.main_streams.packinfo.packsizes == [441]
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [3]
    assert archive.header.main_streams.substreamsinfo.digestsdefined == [
        True,
        True,
        True,
    ]
    assert archive.header.main_streams.substreamsinfo.digests == [
        3010113243,
        3703540999,
        2164028094,
    ]
    assert archive.header.main_streams.substreamsinfo.unpacksizes == [111, 58, 559]
    assert len(archive.header.main_streams.unpackinfo.folders) == 1
    assert len(archive.header.main_streams.unpackinfo.folders[0].coders) == 2
    assert archive.header.main_streams.unpackinfo.numfolders == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid
    assert len(archive.header.main_streams.unpackinfo.folders[0].bindpairs) == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid is True
    assert archive.header.main_streams.unpackinfo.folders[0].unpacksizes == [
        728,
        728,
    ]  # 728 = 111 + 58 + 559
    assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
    assert archive.header.main_streams.unpackinfo.folders[0].crc is None
    archive._fpclose()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    m = hashlib.sha256()
    m.update((tmp_path / "tgt" / "setup.py").open("rb").read())
    assert m.digest() == binascii.unhexlify("b916eed2a4ee4e48c51a2b51d07d450de0be4dbb83d20e67f6fd166ff7921e49")
    m = hashlib.sha256()
    m.update((tmp_path / "tgt" / "scripts" / "py7zr").open("rb").read())
    assert m.digest() == binascii.unhexlify("b0385e71d6a07eb692f5fb9798e9d33aaf87be7dfff936fd2473eab2a593d4fd")
    dc = filecmp.dircmp(tmp_path.joinpath("src"), tmp_path.joinpath("tgt"))
    assert dc.diff_files == []
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_file_0(capsys, tmp_path):
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(os.path.join(testdata_path, "test1.txt"), "test1.txt")
    assert len(archive.files) == 1
    archive.close()
    with target.open("rb") as target_archive:
        val = target_archive.read(1000)
        assert val.startswith(py7zr.properties.MAGIC_7Z)
    archive = py7zr.SevenZipFile(target, "r")
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders[0] == 1
    assert archive.testzip() is None
    mtime = datetime.fromtimestamp(pathlib.Path(os.path.join(testdata_path, "test1.txt")).stat().st_mtime, timezone.utc)
    expected = (
        "total 1 files and directories in archive\n"
        "   Date      Time    Attr         Size   Compressed  Name\n"
        "------------------- ----- ------------ ------------  ------------------------\n"
    )
    expected += "{} ....A           33           37  test1.txt\n".format(ltime(mtime))
    expected += "------------------- ----- ------------ ------------  ------------------------\n"
    cli = py7zr.cli.Cli()
    cli.run(["l", str(target)])
    out, err = capsys.readouterr()
    archive.close()
    assert expected == out


@pytest.mark.basic
def test_compress_directory(tmp_path):
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    assert len(archive.files) == 2
    archive._write_flush()
    assert archive.header.main_streams.packinfo.numstreams == 1
    assert archive.header.main_streams.packinfo.packsizes == [17]
    assert archive.header.main_streams.unpackinfo.numfolders == 1
    assert len(archive.header.main_streams.unpackinfo.folders) == 1
    assert len(archive.header.main_streams.unpackinfo.folders[0].coders) == 2
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.substreamsinfo.unpacksizes == [11]
    assert len(archive.header.files_info.files) == 2
    archive._fpclose()
    with target.open("rb") as target_archive:
        val = target_archive.read(1000)
        assert val.startswith(py7zr.properties.MAGIC_7Z)
    archive = py7zr.SevenZipFile(target, "r")
    assert archive.testzip() is None
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_files_1(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_1.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(".")
    archive._write_flush()
    assert len(archive.files) == 4
    assert len(archive.header.files_info.files) == 4
    expected = [True, False, False, False]
    for i, f in enumerate(archive.header.files_info.files):
        f["emptystream"] = expected[i]
    assert archive.header.files_info.emptyfiles == [True, False, False, False]
    assert archive.header.files_info.files[3]["emptystream"] is False
    expected_attributes = stat.FILE_ATTRIBUTE_ARCHIVE
    if os.name == "posix":
        expected_attributes |= 0x8000 | (0o644 << 16)
    assert archive.header.files_info.files[3]["attributes"] == expected_attributes
    assert archive.header.files_info.files[3]["maxsize"] == 441
    assert archive.header.files_info.files[3]["uncompressed"] == 559
    assert archive.header.files_info.files[1]["filename"] == "scripts/py7zr"
    assert archive.header.files_info.files[3]["filename"] == "setup.py"
    assert archive.header.main_streams.packinfo.numstreams == 1
    assert archive.header.main_streams.packinfo.packsizes == [441]
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [3]
    assert archive.header.main_streams.substreamsinfo.digestsdefined == [
        True,
        True,
        True,
    ]
    assert archive.header.main_streams.substreamsinfo.digests == [
        3010113243,
        3703540999,
        2164028094,
    ]
    assert archive.header.main_streams.substreamsinfo.unpacksizes == [111, 58, 559]
    assert len(archive.header.main_streams.unpackinfo.folders) == 1
    assert len(archive.header.main_streams.unpackinfo.folders[0].coders) == 2
    assert archive.header.main_streams.unpackinfo.numfolders == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid
    assert len(archive.header.main_streams.unpackinfo.folders[0].bindpairs) == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid is True
    assert archive.header.main_streams.unpackinfo.folders[0].unpacksizes == [
        728,
        728,
    ]  # 728 = 111 + 58 + 559
    assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
    assert archive.header.main_streams.unpackinfo.folders[0].crc is None
    archive._fpclose()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    m = hashlib.sha256()
    m.update((tmp_path / "tgt" / "setup.py").open("rb").read())
    assert m.digest() == binascii.unhexlify("b916eed2a4ee4e48c51a2b51d07d450de0be4dbb83d20e67f6fd166ff7921e49")
    m = hashlib.sha256()
    m.update((tmp_path / "tgt" / "scripts" / "py7zr").open("rb").read())
    assert m.digest() == binascii.unhexlify("b0385e71d6a07eb692f5fb9798e9d33aaf87be7dfff936fd2473eab2a593d4fd")
    dc = filecmp.dircmp(tmp_path.joinpath("src"), tmp_path.joinpath("tgt"))
    assert dc.diff_files == []
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        assert archive.test() is None
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.api
def test_register_archive_format(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    # Prepare test data
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_1.7z"), path=tmp_path.joinpath("src"))
    #
    shutil.register_archive_format("7zip", pack_7zarchive, description="7zip archive")
    shutil.make_archive(str(tmp_path.joinpath("target")), "7zip", str(tmp_path.joinpath("src")))
    # check result
    archive = SevenZipFile(tmp_path.joinpath("target.7z"), "r")
    archive.extractall(path=tmp_path.joinpath("tgt"))
    archive.close()
    m = hashlib.sha256()
    m.update((tmp_path / "tgt" / "setup.py").open("rb").read())
    assert m.digest() == binascii.unhexlify("b916eed2a4ee4e48c51a2b51d07d450de0be4dbb83d20e67f6fd166ff7921e49")
    m = hashlib.sha256()
    m.update((tmp_path / "tgt" / "scripts" / "py7zr").open("rb").read())
    assert m.digest() == binascii.unhexlify("b0385e71d6a07eb692f5fb9798e9d33aaf87be7dfff936fd2473eab2a593d4fd")


@pytest.mark.api
def test_compress_with_simple_filter(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.api
def test_compress_with_custom_filter(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_DELTA, "dist": 5},
        {"id": py7zr.FILTER_LZMA2, "preset": 7 | py7zr.PRESET_EXTREME},
    ]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.set_encoded_header_mode(True)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    archive.header.main_streams.packinfo.enable_digests = True
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as arc:
        assert arc.test()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_files_2(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_2.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(str(tmp_path.joinpath("src")))
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(".")
    archive.close()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    dc = filecmp.dircmp(tmp_path.joinpath("src"), tmp_path.joinpath("tgt"))
    assert dc.diff_files == []
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
def test_compress_files_3(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_3.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(".")
    archive.close()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    dc = filecmp.dircmp(tmp_path.joinpath("src"), tmp_path.joinpath("tgt"))
    assert dc.diff_files == []
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
def test_compress_symlink(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "symlink.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    filters = [
        {"id": lzma.FILTER_LZMA2, "preset": 7 | lzma.PRESET_EXTREME},
    ]
    archive = py7zr.SevenZipFile(target, "w", filters=filters)
    archive.set_encoded_header_mode(False)
    archive.writeall(".")
    archive._write_flush()
    assert len(archive.header.files_info.files) == 6
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [5]
    assert len(archive.files) == 6
    assert len(archive.header.files_info.files) == 6
    expected = [True, False, False, False, False, False]
    for i, f in enumerate(archive.header.files_info.files):
        f["emptystream"] = expected[i]
    assert archive.header.files_info.files[5]["maxsize"] == 1543
    assert archive.header.main_streams.packinfo.packsizes == [1543]
    assert archive.header.files_info.files[4]["uncompressed"] == 6536
    assert archive.header.files_info.files[1]["filename"] == "lib/libabc.so"
    assert archive.header.files_info.files[2]["filename"] == "lib/libabc.so.1"
    if os.name == "nt":
        assert check_bit(
            archive.header.files_info.files[2]["attributes"],
            stat.FILE_ATTRIBUTE_REPARSE_POINT,
        )
    else:
        assert check_bit(
            archive.header.files_info.files[2]["attributes"],
            FILE_ATTRIBUTE_UNIX_EXTENSION,
        )
        assert stat.S_ISLNK(archive.header.files_info.files[2]["attributes"] >> 16)
    assert archive.header.main_streams.packinfo.numstreams == 1
    assert archive.header.main_streams.substreamsinfo.digestsdefined == [
        True,
        True,
        True,
        True,
        True,
    ]
    assert archive.header.main_streams.substreamsinfo.unpacksizes == [
        11,
        13,
        15,
        6536,
        3,
    ]
    assert archive.header.main_streams.substreamsinfo.digests == [
        4262439050,
        2607345479,
        2055456646,
        437637236,
        2836347852,
    ]
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [5]
    assert len(archive.header.main_streams.unpackinfo.folders) == 1
    assert len(archive.header.main_streams.unpackinfo.folders[0].coders) == 1
    assert archive.header.main_streams.unpackinfo.numfolders == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid
    assert archive.header.main_streams.unpackinfo.folders[0].bindpairs == []
    assert archive.header.main_streams.unpackinfo.folders[0].solid is True
    assert archive.header.main_streams.unpackinfo.folders[0].unpacksizes == [6578]
    assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
    assert archive.header.main_streams.unpackinfo.folders[0].crc is None
    archive._fpclose()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()


@pytest.mark.files
def test_compress_zerofile(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    with tmp_path.joinpath("src", "f").open(mode="w") as f:
        f.write("")
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(".")
    archive._write_flush()
    assert len(archive.header.files_info.files) == 1
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [1]
    assert len(archive.files) == 1
    assert len(archive.header.files_info.files) == 1
    expected = [True]
    for i, f in enumerate(archive.header.files_info.files):
        f["emptystream"] = expected[i]
    archive._fpclose()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_directories(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt1").mkdir()
    tmp_path.joinpath("tgt2").mkdir()
    # target files
    tmp_path.joinpath("src", "dir1").mkdir()
    tmp_path.joinpath("src", "dir2").mkdir()
    tmp_path.joinpath("src", "dir3").mkdir()
    tmp_path.joinpath("src", "dir4").mkdir()
    tmp_path.joinpath("src", "dir5").mkdir()
    tmp_path.joinpath("src", "dir6").mkdir()
    tmp_path.joinpath("src", "dir7").mkdir()
    tmp_path.joinpath("src", "dir8").mkdir()
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(".")
    archive._write_flush()
    for i, f in enumerate(archive.header.files_info.files):
        f["emptystream"] = True
    archive._fpclose()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt1"))
    reader.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
def test_compress_absolute_symlink_as_relative(tmp_path):
    # test case derived by github issue#112
    src = tmp_path / "src"
    # create symbolic link and actual file
    src.mkdir()
    src.joinpath("rel").mkdir()
    origin = src / "Original1.txt"
    with origin.open("w") as f:
        f.write("Original1.txt")
    s = src / "rel" / "link_to_Original1.txt"
    s.symlink_to(origin, False)
    # create archive with symlink
    target = tmp_path / "symb_2.7z"
    with py7zr.SevenZipFile(target, "w") as archive:
        for fname, arcname in [
            (origin, "Original1.txt"),
            (s.parent, "rel"),
            (s, "rel/link_to_Original1.txt"),
        ]:
            archive.write(fname, arcname)
    # extract archive
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path.joinpath("tgt"))
    # check link
    tpath = py7zr.helpers.readlink(tmp_path / "tgt" / "rel" / "link_to_Original1.txt")
    assert pathlib.Path(tpath).as_posix() == "../Original1.txt"


@pytest.mark.files
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
def test_compress_files_deref(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "symlink.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(tmp_path.joinpath("src"))
    with py7zr.SevenZipFile(target, "w", dereference=True) as archive:
        archive.writeall(".")
    with py7zr.SevenZipFile(target, "r") as reader:
        reader.extractall(path=tmp_path.joinpath("tgt"))
    assert tmp_path.joinpath("tgt").joinpath("lib64").is_dir()
    assert tmp_path.joinpath("tgt").joinpath("lib/libabc.so").is_file()
    assert tmp_path.joinpath("tgt").joinpath("lib64/libabc.so").is_file()


@pytest.mark.basic
def test_compress_lzma2_bcj(tmp_path):
    my_filters = [{"id": py7zr.FILTER_X86}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "lzma2bcj.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(tmp_path / "src", "src")
    archive._write_flush()
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [12]
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid
    assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].incoder == 1
    assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].outcoder == 0
    assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
    assert archive.header.main_streams.unpackinfo.folders[0].crc is None
    archive._fpclose()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [12]
        assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numinstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numoutstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].incoder == 1
        assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].outcoder == 0
        assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
        assert archive.header.main_streams.unpackinfo.folders[0].crc is None
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
@pytest.mark.skipif(
    sys.platform.startswith("win") and sys.version_info < (3, 7),
    reason="requires python3.7 or higher for windows",
)
def test_compress_files_deref_loop(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "symlink.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    os.chdir(str(tmp_path.joinpath("src")))
    # create symlink loop
    tmp_path.joinpath("src/lib/parent").symlink_to(tmp_path.joinpath("src/lib"), target_is_directory=True)
    with py7zr.SevenZipFile(target, "w", dereference=True) as archive:
        archive.writeall(".")


@pytest.mark.basic
def test_compress_writestr1(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    data = b"this is data"
    with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.writestr(data, "src.txt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_writestr2(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    data = "this is data"
    with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.writestr(data, "src.txt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_writef1(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    data = b"this is data"
    with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.writef(io.BytesIO(data), "src.txt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_writef2(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    with open(os.path.join(testdata_path, "test1.txt"), "rb") as src:
        with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
            archive.writef(src, "test1.txt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_writef3(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    with pytest.raises(ValueError):
        with open(os.path.join(testdata_path, "test1.txt"), "r") as src:
            with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
                archive.writef(src, "test1.txt")


@pytest.mark.basic
def test_compress_copy(tmp_path):
    my_filters = [{"id": py7zr.FILTER_COPY}]
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "lzma2bcj.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(tmp_path / "src", "src")
    archive._write_flush()
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [12]
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid
    assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
    assert archive.header.main_streams.unpackinfo.folders[0].crc is None
    archive._fpclose()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_multi_filter_delta(tmp_path):
    my_filters = [{"id": py7zr.FILTER_DELTA}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    assert archive.files is not None
    assert len(archive.files) == 2
    for f in archive.files:
        assert f.filename in ("src", pathlib.Path("src").joinpath("bra.txt").as_posix())
    archive.set_encoded_header_mode(False)
    archive._write_flush()
    assert len(archive.header.files_info.files) == 2
    assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [1]
    assert len(archive.header.files_info.files) == 2
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numinstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numoutstreams"] == 1
    assert archive.header.main_streams.unpackinfo.folders[0].solid
    assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].incoder == 1
    assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].outcoder == 0
    assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
    assert archive.header.main_streams.unpackinfo.folders[0].crc is None
    expected = [True, False]
    for i, f in enumerate(archive.header.files_info.files):
        f["emptystream"] = expected[i]
    archive._fpclose()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        assert len(archive.header.files_info.files) == 2
        assert archive.header.main_streams.substreamsinfo.num_unpackstreams_folders == [1]
        assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numinstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].coders[0]["numoutstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numinstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].coders[1]["numoutstreams"] == 1
        assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].incoder == 1
        assert archive.header.main_streams.unpackinfo.folders[0].bindpairs[0].outcoder == 0
        assert archive.header.main_streams.unpackinfo.folders[0].digestdefined is False
        assert archive.header.main_streams.unpackinfo.folders[0].crc is None
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_deflate(tmp_path):
    my_filters = [{"id": py7zr.FILTER_DEFLATE}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_deflate_bcj(tmp_path):
    my_filters = [{"id": py7zr.FILTER_X86}, {"id": py7zr.FILTER_DEFLATE}]
    tmp_path.joinpath("src").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "lzma2bcj.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(tmp_path.joinpath("src"), "src")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    libarchive_extract(target, tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_bz2_bcj(tmp_path):
    my_filters = [{"id": py7zr.FILTER_X86}, {"id": py7zr.FILTER_BZIP2}]
    tmp_path.joinpath("src").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "lzma2bcj.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(tmp_path.joinpath("src"), "src")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    libarchive_extract(target, tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_zstd(tmp_path):
    my_filters = [{"id": py7zr.FILTER_ZSTD}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    libarchive_extract(target, tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_zstd_2(tmp_path):
    my_filters = [{"id": py7zr.FILTER_ZSTD}]
    tmp_path.joinpath("src").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "lzma2bcj.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(tmp_path.joinpath("src"), "src")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")


@pytest.mark.basic
def test_compress_arm64(tmp_path):
    my_filters = [{"id": py7zr.FILTER_ARM}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.write(os.path.join(testdata_path, "lib"), "lib")
    archive.writeall(os.path.join(testdata_path, "lib", "aarch64-linux-gnu"), "lib/aarch64-linux-gnu")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_ppc(tmp_path):
    my_filters = [{"id": py7zr.FILTER_POWERPC}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.write(os.path.join(testdata_path, "lib"), "lib")
    archive.writeall(
        os.path.join(testdata_path, "lib", "powerpc64le-linux-gnu"),
        "lib/powerpc64le-linux-gnu",
    )
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_ia64(tmp_path):
    my_filters = [{"id": py7zr.FILTER_IA64}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.write(os.path.join(testdata_path, "lib"), "lib")
    archive.writeall(
        os.path.join(testdata_path, "lib", "powerpc64le-linux-gnu"),
        "lib/powerpc64le-linux-gnu",
    )
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_sparc(tmp_path):
    my_filters = [{"id": py7zr.FILTER_SPARC}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.write(os.path.join(testdata_path, "lib"), "lib")
    archive.writeall(
        os.path.join(testdata_path, "lib", "powerpc64le-linux-gnu"),
        "lib/powerpc64le-linux-gnu",
    )
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_armt(tmp_path):
    my_filters = [
        {"id": py7zr.FILTER_ARMTHUMB},
        {"id": py7zr.FILTER_LZMA2, "preset": 7},
    ]
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.write(os.path.join(testdata_path, "lib"), "lib")
    archive.writeall(os.path.join(testdata_path, "lib", "aarch64-linux-gnu"), "lib/aarch64-linux-gnu")
    archive.close()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_small_files(tmp_path):
    tmp_path.joinpath("t").mkdir()
    with tmp_path.joinpath("t/a").open("w") as f:
        f.write("1")
    with tmp_path.joinpath("t/b").open("w") as f:
        f.write("2")
    with py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), "w") as archive:
        archive.writeall(tmp_path.joinpath("t"), "t")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_append(tmp_path):
    target = tmp_path.joinpath("target.7z")
    shutil.copy(os.path.join(testdata_path, "test_1.7z"), target)
    archive = py7zr.SevenZipFile(target, mode="a")
    archive.encoded_header_mode = False
    archive.write(os.path.join(testdata_path, "test1.txt"), "test1.txt")
    archive._write_flush()
    archive._fpclose()
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_append_2(tmp_path):
    target = tmp_path.joinpath("target.7z")
    shutil.copy(os.path.join(testdata_path, "lzma2bcj.7z"), target)
    with py7zr.SevenZipFile(target, mode="a") as archive:
        archive.write(os.path.join(testdata_path, "test1.txt"), "test1.txt")
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path / "tgt")
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_append_files_2(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_2.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    shutil.copy(os.path.join(testdata_path, "test_1.7z"), target)
    os.chdir(str(tmp_path.joinpath("src")))
    with py7zr.SevenZipFile(target, "a") as archive:
        archive.writeall(".")
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.test()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_append_empty_files(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    with tmp_path.joinpath("src").joinpath("1.txt").open(mode="w") as w:
        pass
    with tmp_path.joinpath("src").joinpath("2.txt").open(mode="w") as w:
        pass
    target = tmp_path.joinpath("target.7z")
    os.chdir(str(tmp_path.joinpath("src")))
    with py7zr.SevenZipFile(target, "w") as archive:
        archive.write("1.txt")
    with py7zr.SevenZipFile(target, "a") as archive:
        archive.write("2.txt")
    #
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.test()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_archive_emptyfile_1(tmp_path):
    tmp_path.joinpath("src").mkdir()
    with tmp_path.joinpath("src", "x").open(mode="wb") as f:
        f.write(b"")
    archive = py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), "w")
    archive.write(tmp_path.joinpath("src", "x"), "y")
    archive.close()
    #
    with py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), "r") as arc:
        arc.extractall(path=tmp_path / "tgt")
    #
    assert tmp_path.joinpath("tgt", "y").is_file()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.basic
def test_compress_null_archive(tmp_path):
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    assert len(archive.files) == 0
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))
    #
    with py7zr.SevenZipFile(target, "r") as arc:
        arc.extractall(path=tmp_path / "tgt")


@pytest.mark.basic
def test_compress_append_archive(tmp_path):
    target = tmp_path.joinpath("target.7z")
    with py7zr.SevenZipFile(target, "w"):
        pass
    archive = py7zr.SevenZipFile(target, "a")
    archive.writeall(os.path.join(testdata_path, "test1.txt"), "test1.txt")
    assert len(archive.files) == 1
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))
    #
    with py7zr.SevenZipFile(target, "r") as arc:
        arc.extractall(path=tmp_path / "tgt")


@pytest.mark.basic
def test_compress_append_writestr_archive(tmp_path):
    target = tmp_path.joinpath("target.7z")
    with py7zr.SevenZipFile(target, "w"):
        pass
    archive = py7zr.SevenZipFile(target, "a")
    archive.set_encoded_header_mode(False)
    archive.writestr("Test", arcname="Test/Test.txt")
    assert len(archive.files.files_list) == 1
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))
    #
    with py7zr.SevenZipFile(target, "r") as arc:
        arc.extractall(path=tmp_path / "tgt")


@pytest.mark.basic
def test_compress_simple_file_0(tmp_path):
    with tmp_path.joinpath("target.7z").open(mode="wb") as target:
        with py7zr.SevenZipFile(target, "w") as archive:
            archive.writeall(os.path.join(testdata_path, "src"), "src")


@pytest.mark.basic
def test_compress_append_archive_w_zerofile(tmp_path):
    tmp_path.joinpath("src").mkdir()
    with tmp_path.joinpath("src", "f").open(mode="w") as f:
        f.write("")
    target = tmp_path.joinpath("target.7z")
    with py7zr.SevenZipFile(target, "w") as archive:
        archive.write(os.path.join(testdata_path, "test1.txt"), "test1.txt")
        archive.write(tmp_path.joinpath("src", "f"), "f")
    with py7zr.SevenZipFile(target, "a") as archive:
        archive.writestr("Test", arcname="Test/Test.txt")
        assert len(archive.files.files_list) == 3
    archive = py7zr.SevenZipFile(target, "a")
    archive.set_encoded_header_mode(False)
    assert len(archive.files.files_list) == 3
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))
    #
    with py7zr.SevenZipFile(target, "r") as arc:
        arc.extractall(path=tmp_path / "tgt")


@pytest.mark.files
def test_compress_absolute_arcname(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_2.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(str(tmp_path.joinpath("src")), "/foo")
    archive.close()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    dc = filecmp.dircmp(tmp_path.joinpath("src"), tmp_path.joinpath("tgt").joinpath("foo"))
    assert dc.diff_files == []
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_compress_win32_absolute_arcname(tmp_path):
    tmp_path.joinpath("src").mkdir()
    tmp_path.joinpath("tgt").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "test_2.7z"), path=tmp_path.joinpath("src"))
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w")
    archive.set_encoded_header_mode(False)
    archive.writeall(str(tmp_path.joinpath("src")), "C:/foo")
    archive.close()
    reader = py7zr.SevenZipFile(target, "r")
    reader.extractall(path=tmp_path.joinpath("tgt"))
    reader.close()
    dc = filecmp.dircmp(tmp_path.joinpath("src"), tmp_path.joinpath("tgt").joinpath("foo"))
    assert dc.diff_files == []
    #
    p7zip_test(tmp_path / "target.7z")
    libarchive_extract(tmp_path / "target.7z", tmp_path.joinpath("tgt2"))


@pytest.mark.files
def test_empty_stream(tmp_path):
    archive = py7zr.SevenZipFile(file=tmp_path / "test.7z", mode="w")
    archive.writestr(data="", arcname="empty.txt")
    assert len(archive.files.files_list) == 1
    assert archive.header.files_info.files[0]["uncompressed"] == 0
    archive.close()
    p7zip_test(tmp_path / "test.7z")



import binascii
import ctypes
import datetime
import hashlib
import io
import lzma
import os
import pathlib
import stat
import struct
import sys

import pytest

import py7zr.archiveinfo
import py7zr.compressor
import py7zr.helpers
import py7zr.io
import py7zr.properties
from py7zr.py7zr import FILE_ATTRIBUTE_UNIX_EXTENSION

testdata_path = os.path.join(os.path.dirname(__file__), "data")


@pytest.mark.unit
def test_py7zr_signatureheader():
    header_data = io.BytesIO(
        b"\x37\x7a\xbc\xaf\x27\x1c\x00\x02\x70\x2a\xb7\x37\xa0\x00\x00\x00\x00\x00\x00\x00\x21"
        b"\x00\x00\x00\x00\x00\x00\x00\xb9\xb8\xe4\xbf"
    )
    header = py7zr.archiveinfo.SignatureHeader.retrieve(header_data)
    assert header is not None
    assert header.version == (b"\x00", b"\x02")
    assert header.nextheaderofs == 160


@pytest.mark.unit
def test_py7zr_mainstreams():
    header_data = io.BytesIO(
        b"\x04\x06\x00\x01\t0\x00\x07\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00\x00\x00\x02\x0cB\x00"
        b"\x08\r\x02\t!\n\x01>jb\x08\xce\x9a\xb7\x88\x00\x00"
    )
    pid = header_data.read(1)
    assert pid == py7zr.properties.Property.MAIN_STREAMS_INFO
    streams = py7zr.archiveinfo.StreamsInfo.retrieve(header_data)
    assert streams is not None


@pytest.mark.unit
def test_py7zr_folder_retrieve():
    header_data = io.BytesIO(b"\x0b" b"\x01\x00\x01#\x03\x01\x01\x05]\x00\x10\x00\x00")
    pid = header_data.read(1)
    assert pid == py7zr.properties.Property.FOLDER
    num_folders = py7zr.archiveinfo.read_byte(header_data)
    assert num_folders == 1
    external = py7zr.archiveinfo.read_byte(header_data)
    assert external == 0x00
    folder = py7zr.archiveinfo.Folder.retrieve(header_data)
    assert folder.packed_indices == [0]
    assert folder.digestdefined is False
    coder = folder.coders[0]
    assert coder["method"] == b"\x03\x01\x01"
    assert coder["properties"] == b"]\x00\x10\x00\x00"
    assert coder["numinstreams"] == 1
    assert coder["numoutstreams"] == 1


@pytest.mark.unit
def test_py7zr_folder_write():
    folders = []
    for _ in range(1):
        folder = py7zr.archiveinfo.Folder()
        folder.bindpairs = []
        folder.coders = [
            {
                "method": b"\x03\x01\x01",
                "numinstreams": 1,
                "numoutstreams": 1,
                "properties": b"]\x00\x10\x00\x00",
            }
        ]
        folder.crc = None
        folder.digestdefined = False
        folder.packed_indices = [0]
        folder.solid = True
        folders.append(folder)
    #
    buffer = io.BytesIO()
    # following should be run in StreamsInfo class.
    py7zr.archiveinfo.write_byte(buffer, py7zr.properties.Property.FOLDER)
    py7zr.archiveinfo.write_uint64(buffer, len(folders))
    external = b"\x00"
    py7zr.archiveinfo.write_byte(buffer, external)
    for folder in folders:
        folder.write(buffer)
    actual = buffer.getvalue()
    assert actual == b"\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00\x10\x00\x00"


@pytest.mark.unit
def test_py7zr_unpack_info():
    # prepare for unpack_info values
    unpack_info = py7zr.archiveinfo.UnpackInfo()
    unpack_info.folders = []
    for _ in range(1):
        folder = py7zr.archiveinfo.Folder()
        folder.bindpairs = []
        folder.coders = [
            {
                "method": b"\x03\x01\x01",
                "numinstreams": 1,
                "numoutstreams": 1,
                "properties": b"]\x00\x10\x00\x00",
            }
        ]
        folder.crc = None
        folder.digestdefined = False
        folder.packed_indices = [0]
        folder.solid = True
        folder.unpacksizes = [0x22]
        unpack_info.folders.append(folder)
    unpack_info.numfolders = len(unpack_info.folders)
    #
    buffer = io.BytesIO()
    unpack_info.write(buffer)
    actual = buffer.getvalue()
    assert actual == b"\x07\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00\x10\x00\x00\x0c\x22\x00"


@pytest.mark.unit
def test_py7zr_substreamsinfo():
    header_data = io.BytesIO(b"\x08\x0d\x03\x09\x6f\x3a\n\x01\xdb\xaej\xb3\x07\x8d\xbf\xdc\xber\xfc\x80\x00")
    pid = header_data.read(1)
    assert pid == py7zr.properties.Property.SUBSTREAMS_INFO
    folders = [py7zr.archiveinfo.Folder()]
    folders[0].unpacksizes = [728]
    numfolders = 1
    ss = py7zr.archiveinfo.SubstreamsInfo.retrieve(header_data, numfolders, folders)
    pos = header_data.tell()
    print(pos)
    assert ss.digestsdefined == [True, True, True]
    assert ss.digests[0] == 3010113243
    assert ss.digests[1] == 3703540999
    assert ss.digests[2] == 2164028094
    assert ss.num_unpackstreams_folders[0] == 3
    assert ss.unpacksizes == [111, 58, 559]


@pytest.mark.unit
def test_py7zr_substreamsinfo_write():
    folders = [py7zr.archiveinfo.Folder()]
    folders[0].unpacksizes = [728]
    ss = py7zr.archiveinfo.SubstreamsInfo()
    buffer = io.BytesIO()
    ss.digestsdefined = [True, True, True]
    ss.digests = [3010113243, 3703540999, 2164028094]
    ss.num_unpackstreams_folders = [3]
    ss.unpacksizes = [111, 58, 559]
    ss.write(buffer)
    actual = buffer.getvalue()
    assert actual == b"\x08\x0d\x03\x09\x6f\x3a\n\x01\xdb\xaej\xb3\x07\x8d\xbf\xdc\xber\xfc\x80\x00"


@pytest.mark.unit
def test_py7zr_header():
    fp = open(os.path.join(testdata_path, "solid.7z"), "rb")
    header_data = io.BytesIO(
        b"\x01"
        b"\x04\x06\x00\x01\t0\x00\x07\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00\x00\x00\x02\x0cB\x00"
        b"\x08\r\x02\t!\n\x01>jb\x08\xce\x9a\xb7\x88\x00\x00"
        b"\x05\x03\x0e\x01\x80"
        b"\x11=\x00t\x00e\x00s\x00t\x00\x00\x00"
        b"t\x00e\x00s\x00t\x001\x00.\x00t\x00x\x00t\x00\x00\x00"
        b"t\x00e\x00s\x00t\x00/\x00t\x00e\x00s\x00t\x002\x00.\x00t\x00x\x00t\x00\x00\x00"
        b"\x14\x1a\x01\x00\x04>\xe6\x0f{H\xc6\x01d\xca \x8byH\xc6\x01\x8c\xfa\xb6\x83yH\xc6\x01"
        b"\x15\x0e\x01\x00\x10\x00\x00\x00 \x00\x00\x00 \x00\x00\x00"
        b"\x00\x00"
    )
    header = py7zr.archiveinfo.Header.retrieve(fp, header_data, start_pos=32)
    assert header is not None
    assert header.files_info is not None
    assert header.main_streams is not None
    assert len(header.files_info.files) == 3


@pytest.mark.unit
def test_py7zr_encoded_header():
    fp = open(os.path.join(testdata_path, "test_5.7z"), "rb")
    # set test data to buffer that start with Property.ENCODED_HEADER
    buffer = io.BytesIO(
        b"\x17\x060\x01\tp\x00\x07\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00"
        b"\x00\x10\x00\x0c\x80\x9d\n\x01\xe5\xa1\xb7b\x00\x00"
    )
    header = py7zr.archiveinfo.Header.retrieve(fp, buffer, start_pos=32)
    assert header is not None
    assert header.files_info is not None
    assert header.main_streams is not None
    assert len(header.files_info.files) == 3


@pytest.mark.unit
def test_py7zr_files_info_1():
    header_data = io.BytesIO(
        b"\x05\x03\x0e\x01\x80\x11=\x00t\x00e\x00s\x00t\x00\x00\x00t\x00e\x00s\x00t\x001\x00."
        b"\x00t\x00x\x00t\x00\x00\x00t\x00e\x00s\x00t\x00/\x00t\x00e\x00s\x00t\x002\x00.\x00t\x00x"
        b"\x00t\x00\x00\x00\x14\x1a\x01\x00\x04>\xe6\x0f{H\xc6\x01d\xca \x8byH\xc6\x01\x8c\xfa\xb6"
        b"\x83yH\xc6\x01\x15\x0e\x01\x00\x10\x00\x00\x00 \x00\x00\x00 \x00\x00\x00\x00\x00"
    )
    pid = header_data.read(1)
    assert pid == py7zr.properties.Property.FILES_INFO
    files_info = py7zr.archiveinfo.FilesInfo.retrieve(header_data)
    assert files_info is not None
    assert files_info.files[0].get("filename") == "test"
    assert files_info.files[1].get("filename") == "test1.txt"
    assert files_info.files[2].get("filename") == "test/test2.txt"


@pytest.mark.unit
def test_py7zr_files_info_2():
    header_data = io.BytesIO(
        b"\x05\x04\x11_\x00c\x00o\x00p\x00y\x00i\x00n\x00g\x00.\x00t\x00x\x00t\x00\x00\x00H\x00"
        b"i\x00s\x00t\x00o\x00r\x00y\x00.\x00t\x00x\x00t\x00\x00\x00L\x00i\x00c\x00e\x00n\x00s"
        b"\x00e\x00.\x00t\x00x\x00t\x00\x00\x00r\x00e\x00a\x00d\x00m\x00e\x00.\x00t\x00x\x00t\x00"
        b'\x00\x00\x14"\x01\x00\x00[\x17\xe6\xc70\xc1\x01\x00Vh\xb5\xda\xf8\xc5\x01\x00\x97\xbd'
        b"\xf9\x07\xf7\xc4\x01\x00gK\xa8\xda\xf8\xc5\x01\x15\x12\x01\x00  \x00\x00  \x00"
        b"\x00  \x00\x00  \x00\x00\x00\x00"
    )
    pid = header_data.read(1)
    assert pid == py7zr.properties.Property.FILES_INFO
    files_info = py7zr.archiveinfo.FilesInfo.retrieve(header_data)
    assert files_info is not None
    assert len(files_info.files) == 4
    assert files_info.files[0].get("filename") == "copying.txt"
    assert files_info.files[0].get("attributes") == 0x2020
    assert files_info.files[1].get("filename") == "History.txt"
    assert files_info.files[1].get("attributes") == 0x2020
    assert files_info.files[2].get("filename") == "License.txt"
    assert files_info.files[2].get("attributes") == 0x2020
    assert files_info.files[3].get("filename") == "readme.txt"
    assert files_info.files[3].get("attributes") == 0x2020


@pytest.mark.unit
def test_lzma_lzma2_compressor():
    filters = [{"id": 33, "dict_size": 16777216}]
    assert lzma.LZMADecompressor(format=lzma.FORMAT_RAW, filters=filters) is not None


@pytest.mark.unit
def test_lzma_lzma2bcj_compressor():
    filters = [{"id": 4}, {"id": 33, "dict_size": 16777216}]
    assert lzma.LZMADecompressor(format=lzma.FORMAT_RAW, filters=filters) is not None


@pytest.mark.unit
@pytest.mark.parametrize(
    "abytes, count, checkall, expected",
    [
        (
            b"\xb4\x80",
            9,
            False,
            [True, False, True, True, False, True, False, False, True],
        ),
        (
            b"\xff\xc0",
            10,
            False,
            [True, True, True, True, True, True, True, True, True, True],
        ),
        (
            b"\xff\xff\x80",
            17,
            False,
            [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
        ),
    ],
)
def test_read_booleans(abytes, count, checkall, expected):
    buf = io.BytesIO(abytes)
    actual = py7zr.archiveinfo.read_boolean(buf, count, checkall)
    assert actual == expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "booleans, all_defined, expected",
    [
        (
            [True, False, True, True, False, True, False, False, True],
            False,
            b"\xb4\x80",
        ),
        (
            [True, True, True, True, True, True, True, True, True, True],
            False,
            b"\xff\xc0",
        ),
        (
            [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ],
            False,
            b"\xff\xff\x80",
        ),
        (
            [True, False, True, True, False, True, False, False, True],
            True,
            b"\x00\xb4\x80",
        ),
    ],
)
def test_write_booleans(booleans, all_defined, expected):
    buffer = io.BytesIO()
    py7zr.archiveinfo.write_boolean(buffer, booleans, all_defined=all_defined)
    actual = buffer.getvalue()
    assert actual == expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "testinput, expected",
    [
        (1, b"\x01"),
        (127, b"\x7f"),
        (128, b"\x80\x80"),
        (65535, b"\xc0\xff\xff"),
        (441, b"\x81\xb9"),
        (0xFFFF7F, b"\xe0\x7f\xff\xff"),
        (0x0E002100, b"\xee\x00\x21\x00"),
        (0xFFFFFFFF, b"\xf0\xff\xff\xff\xff"),
        (0x7F1234567F, b"\xf8\x7f\x56\x34\x12\x7f"),
        (0x1234567890ABCD, b"\xfe\xcd\xab\x90\x78\x56\x34\x12"),
        (0xCF1234567890ABCD, b"\xff\xcd\xab\x90\x78\x56\x34\x12\xcf"),
    ],
)
def test_write_uint64(testinput, expected):
    buf = io.BytesIO()
    py7zr.archiveinfo.write_uint64(buf, testinput)
    actual = buf.getvalue()
    assert actual == expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "testinput, expected",
    [
        (b"\x01", 1),
        (b"\x7f", 127),
        (b"\x80\x80", 128),
        (b"\x81\xb9", 441),
        (b"\xc0\xff\xff", 65535),
        (b"\xe0\x7f\xff\xff", 0xFFFF7F),
        (b"\xee\x00\x21\x00", 0x0E002100),
        (b"\xf0\xff\xff\xff\xff", 0xFFFFFFFF),
        (b"\xf8\x7f\x56\x34\x12\x7f", 0x7F1234567F),
        (b"\xfe\xcd\xab\x90\x78\x56\x34\x12", 0x1234567890ABCD),
        (b"\xff\xcd\xab\x90\x78\x56\x34\x12\xcf", 0xCF1234567890ABCD),
    ],
)
def test_read_uint64(testinput, expected):
    buf = io.BytesIO(testinput)
    assert py7zr.archiveinfo.read_uint64(buf) == expected


@pytest.mark.unit
@pytest.mark.parametrize("testinput, expected", [(b"t\x00e\x00s\x00t\x00\x00\x00", "test")])
def test_read_utf16(testinput, expected):
    buf = io.BytesIO(testinput)
    actual = py7zr.archiveinfo.read_utf16(buf)
    assert actual == expected


@pytest.mark.unit
@pytest.mark.parametrize("testinput, expected", [("test", b"t\x00e\x00s\x00t\x00\x00\x00")])
def test_write_utf16(testinput, expected):
    buf = io.BytesIO()
    py7zr.archiveinfo.write_utf16(buf, testinput)
    actual = buf.getvalue()
    assert actual == expected


@pytest.mark.unit
def test_write_packinfo():
    packinfo = py7zr.archiveinfo.PackInfo()
    packinfo.packpos = 0
    packinfo.packsizes = [48]
    packinfo.crcs = [py7zr.helpers.calculate_crc32(b"abcd")]
    packinfo.digestdefined = [True]
    packinfo.enable_digests = True
    packinfo.numstreams = 1
    buffer = io.BytesIO()
    packinfo.write(buffer)
    actual = buffer.getvalue()
    assert actual == b"\x06\x00\x01\t0\n\x01\x11\xcd\x82\xed\x00"


@pytest.mark.unit
def test_utc():
    dt = datetime.datetime(2019, 6, 1, 12, 13, 14, 0, tzinfo=py7zr.helpers.UTC())
    assert dt.tzname() == "UTC"


@pytest.mark.unit
def test_utcoffset():
    dt = datetime.datetime(2019, 6, 1, 12, 13, 14, 0)
    assert py7zr.helpers.Local.utcoffset(dt) is not None


@pytest.mark.unit
def test_dst():
    dt = datetime.datetime(2019, 6, 1, 12, 13, 14, 0)
    dst = py7zr.helpers.Local.dst(dt)
    assert dst == datetime.timedelta(0) or dst == datetime.timedelta(hours=1)


@pytest.mark.unit
def test_localtime_tzname():
    dt = datetime.datetime(2019, 6, 1, 12, 13, 14, 0)
    assert py7zr.helpers.Local.tzname(dt) is not None


@pytest.mark.unit
def test_read_crcs():
    """Unit test for read_crcs()"""
    buf = io.BytesIO()
    data0 = 1024
    data1 = 32
    data2 = 1372678730
    buf.write(struct.pack("<L", data0))
    buf.write(struct.pack("<L", data1))
    buf.write(struct.pack("<L", data2))
    buf.seek(0)
    crcs = py7zr.archiveinfo.read_crcs(buf, 3)
    assert crcs[0] == data0
    assert crcs[1] == data1
    assert crcs[2] == data2


@pytest.mark.unit
def test_fileinfo_st_fmt():
    file_info = {}
    file_info["attributes"] = FILE_ATTRIBUTE_UNIX_EXTENSION
    file = py7zr.py7zr.ArchiveFile(0, file_info)
    assert file.st_fmt == 0
    file_info["attributes"] = 0
    file = py7zr.py7zr.ArchiveFile(0, file_info)
    assert file.st_fmt is None


@pytest.mark.unit
def test_wrong_mode():
    with pytest.raises(ValueError):
        py7zr.SevenZipFile(os.path.join(testdata_path, "test_1.7z"), "z")


@pytest.mark.unit
def test_calculate_crc32():
    test_data = b"\x12\x11\x12\x11"
    expected = 3572816238
    assert py7zr.helpers.calculate_crc32(test_data) == expected


@pytest.mark.unit
def test_calculate_crc32_blocksize():
    test_data = open(os.path.join(testdata_path, "test_1.7z"), "rb").read()
    expected = 0x3B04C652
    assert py7zr.helpers.calculate_crc32(test_data, value=0, blocksize=128) == expected


@pytest.mark.unit
def test_startheader_calccrc():
    startheader = py7zr.archiveinfo.SignatureHeader()
    startheader.version = (0, 4)
    startheader.nextheaderofs = 0x000000A0
    startheader.nextheadersize = 0x00000021
    # set test data to buffer that start with Property.ENCODED_HEADER
    header_data = (
        b"\x17\x060\x01\tp\x00\x07\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00"
        b"\x00\x10\x00\x0c\x80\x9d\n\x01\xe5\xa1\xb7b\x00\x00"
    )
    header_crc = py7zr.helpers.calculate_crc32(header_data)
    startheader.calccrc(len(header_data), header_crc)
    assert startheader.nextheadercrc == 0xBFE4B8B9
    assert startheader.startheadercrc == 0x37B72A70


@pytest.mark.unit
def test_write_signature_header():
    startheader = py7zr.archiveinfo.SignatureHeader()
    startheader.nextheaderofs = 0x000000A0
    startheader.nextheadersize = 0x00000021
    header_data = (
        b"\x17\x060\x01\tp\x00\x07\x0b\x01\x00\x01#\x03\x01\x01\x05]\x00"
        b"\x00\x10\x00\x0c\x80\x9d\n\x01\xe5\xa1\xb7b\x00\x00"
    )
    header_crc = py7zr.helpers.calculate_crc32(header_data)
    startheader.calccrc(len(header_data), header_crc)
    file = io.BytesIO()
    startheader.write(file)
    val = file.getvalue()
    assert val.startswith(py7zr.properties.MAGIC_7Z)
    exp_ver = b"\x00\x04"
    exp_scrc = b"\x70\x2a\xb7\x37"
    exp_ofs = b"\xa0\x00\x00\x00\x00\x00\x00\x00"
    exp_size = b"\x21\x00\x00\x00\x00\x00\x00\x00"
    exp_hcrc = b"\xb9\xb8\xe4\xbf"
    assert val == py7zr.properties.MAGIC_7Z + exp_ver + exp_scrc + exp_ofs + exp_size + exp_hcrc


@pytest.mark.unit
def test_make_file_info1():
    file_info = py7zr.py7zr.SevenZipFile._make_file_info(
        pathlib.Path(os.path.join(testdata_path, "src", "bra.txt")), "src/bra.txt"
    )
    assert file_info.get("filename") == "src/bra.txt"
    assert not file_info.get("emptystream")
    assert file_info.get("uncompressed") == 11


@pytest.mark.unit
def test_make_file_info2():
    file_info = py7zr.py7zr.SevenZipFile._make_file_info(pathlib.Path(testdata_path).joinpath("src"))
    assert file_info.get("filename") == pathlib.Path(testdata_path).joinpath("src").as_posix()
    assert file_info.get("emptystream")
    flag = stat.FILE_ATTRIBUTE_DIRECTORY
    assert file_info.get("attributes") & flag == flag


@pytest.mark.unit
def test_simple_compress_and_decompress():
    filters = [
        {"id": lzma.FILTER_LZMA2, "preset": 7 | lzma.PRESET_DEFAULT},
    ]
    lzc = py7zr.compressor.SevenZipCompressor(filters=filters)
    outbuf = io.BytesIO()
    _, _, _ = lzc.compress(io.BytesIO(b"Some data\n"), outbuf)
    _, _, _ = lzc.compress(io.BytesIO(b"Another piece of data\n"), outbuf)
    _, _, _ = lzc.compress(io.BytesIO(b"Even more data\n"), outbuf)
    _ = lzc.flush(outbuf)
    result = outbuf.getvalue()
    size = len(result)
    #
    filters = lzc.filters
    decompressor = lzma.LZMADecompressor(format=lzma.FORMAT_RAW, filters=filters)
    out5 = decompressor.decompress(result)
    assert out5 == b"Some data\nAnother piece of data\nEven more data\n"
    #
    coders = lzc.coders
    crc = py7zr.helpers.calculate_crc32(result)
    decompressor = py7zr.compressor.SevenZipDecompressor(coders, size, [len(out5)], crc)
    outbuf.seek(0, 0)
    out6 = decompressor.decompress(outbuf)
    assert out6 == b"Some data\nAnother piece of data\nEven more data\n"


@pytest.mark.unit
def test_sevenzipcompressor_default():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    indata = io.BytesIO(plain_data)
    outdata = io.BytesIO()
    compressor = py7zr.compressor.SevenZipCompressor(filters=None)
    _, _, _ = compressor.compress(indata, outdata)
    _ = compressor.flush(outdata)
    assert len(outdata.getvalue()) > 1


@pytest.mark.unit
@pytest.mark.parametrize(
    "password, cycle, salt, expected",
    [
        (
            "secret^&",
            0x3F,
            b"i@#ri#Ildajfdk",
            b"i@#ri#Ildajfdks\x00e\x00c\x00r\x00e\x00t\x00^\x00&\x00\x00\x00",
        ),
        (
            "secret",
            5,
            b"",
            b'(6\xa7\xf0\xcc"\t\x9dod\xe2\x8d\xd2\xe9\x08^s\x9c\x99-\xa3N\x08\x10' b"\x9e\\~\x89<\x1cR\xc2",
        ),
    ],
)
def test_calculate_key1(password: str, cycle: int, salt: bytes, expected: bytes):
    key = py7zr.helpers._calculate_key1(password.encode("utf-16LE"), cycle, salt, "sha256")
    assert key == expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "password, cycle, salt, expected",
    [
        (
            "secret^&",
            0x3F,
            b"i@#ri#Ildajfdk",
            b"i@#ri#Ildajfdks\x00e\x00c\x00r\x00e\x00t\x00^\x00&\x00\x00\x00",
        )
    ],
)
def test_calculate_key2(password: str, cycle: int, salt: bytes, expected: bytes):
    key = py7zr.helpers._calculate_key2(password.encode("utf-16LE"), cycle, salt, "sha256")
    assert key == expected


@pytest.mark.unit
@pytest.mark.parametrize(
    "password, cycle, salt, expected",
    [
        (
            "secret^&",
            0x3F,
            b"i@#ri#Ildajfdk",
            b"i@#ri#Ildajfdks\x00e\x00c\x00r\x00e\x00t\x00^\x00&\x00\x00\x00",
        ),
        (
            "secret",
            5,
            b"",
            b'(6\xa7\xf0\xcc"\t\x9dod\xe2\x8d\xd2\xe9\x08^s\x9c\x99-\xa3N\x08\x10' b"\x9e\\~\x89<\x1cR\xc2",
        ),
    ],
)
def test_calculate_key3(password: str, cycle: int, salt: bytes, expected: bytes):
    key = py7zr.helpers._calculate_key3(password.encode("utf-16LE"), cycle, salt, "sha256")
    assert key == expected


def test_calculate_key1_nohash():
    with pytest.raises(ValueError):
        py7zr.helpers._calculate_key1("secret".encode("utf-16LE"), 16, b"", "sha123")


def test_calculate_key2_nohash():
    with pytest.raises(ValueError):
        py7zr.helpers._calculate_key2("secret".encode("utf-16LE"), 16, b"", "sha123")


def test_calculate_key3_nohash():
    with pytest.raises(ValueError):
        py7zr.helpers._calculate_key3("secret".encode("utf-16LE"), 16, b"", "sha123")


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires python3.7")
@pytest.mark.skipif(sys.version_info > (3, 7), reason="requires python3.7")
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
def test_helpers_readlink_dirfd(tmp_path):
    origin = tmp_path / "parent" / "original.txt"
    origin.parent.mkdir(parents=True, exist_ok=True)
    with origin.open("w") as f:
        f.write("Original")
    slink = tmp_path / "target" / "link"
    slink.parent.mkdir(parents=True, exist_ok=True)
    target = pathlib.Path("../parent/original.txt")
    slink.symlink_to(target, False)
    dirfd = os.open(str(origin.parent), os.O_RDONLY | os.O_DIRECTORY)
    assert py7zr.helpers.readlink(slink, dir_fd=dirfd) == target
    os.close(dirfd)


@pytest.mark.unit
def test_compressor_lzma2bcj(tmp_path):
    my_filters = [{"id": py7zr.FILTER_X86}, {"id": py7zr.FILTER_LZMA2, "preset": 7}]
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    indata = io.BytesIO(plain_data)
    outdata = io.BytesIO()
    compressor = py7zr.compressor.SevenZipCompressor(filters=my_filters)
    _, _, _ = compressor.compress(indata, outdata)
    _ = compressor.flush(outdata)
    unpacksizes = compressor.unpacksizes
    outdata.seek(0, 0)
    coders = [
        {"method": b"!", "properties": b"\x18", "numinstreams": 1, "numoutstreams": 1},
        {"method": b"\x03\x03\x01\x03", "numinstreams": 1, "numoutstreams": 1},
    ]
    decompressor = py7zr.compressor.SevenZipDecompressor(
        coders=coders, packsize=len(plain_data), unpacksizes=unpacksizes, crc=None
    )
    revert_data = decompressor.decompress(outdata, max_length=len(plain_data))
    assert revert_data == plain_data


@pytest.mark.unit
def test_lzmadecompressor_lzmabcj():
    indata = b"This file is located in the root."
    compressor = lzma.LZMACompressor(
        format=lzma.FORMAT_RAW,
        filters=[{"id": lzma.FILTER_X86}, {"id": lzma.FILTER_LZMA1}],
    )
    compressed = compressor.compress(indata)
    compressed += compressor.flush()
    decompressor = lzma.LZMADecompressor(
        format=lzma.FORMAT_RAW,
        filters=[{"id": lzma.FILTER_X86}, {"id": lzma.FILTER_LZMA1}],
    )
    outdata = decompressor.decompress(data=compressed)
    assert outdata == indata


@pytest.mark.unit
def test_unit_buffer():
    buf = py7zr.io.Buffer(size=16)
    buf.add(b"12345")
    assert len(buf) == 5
    assert buf.get() == bytearray(b"12345")
    buf.add(b"67890")
    assert buf.view == bytearray(b"67890")
    assert bytes(buf) == b"67890"
    buf.reset()
    assert buf.get() == bytearray(b"")


@pytest.mark.unit
def test_supported_method_is_crypto_id():
    assert py7zr.compressor.SupportedMethods.is_crypto_id(py7zr.properties.FILTER_CRYPTO_AES256_SHA256)


@pytest.mark.unit
def test_supported_method_get_method_id():
    id = py7zr.compressor.SupportedMethods.get_method_id(py7zr.properties.FILTER_CRYPTO_AES256_SHA256)
    assert id == py7zr.properties.CompressionMethod.CRYPT_AES256_SHA256


@pytest.mark.unit
def test_bcj_encode(tmp_path):
    with open(os.path.join(testdata_path, "x86.bin"), "rb") as f:
        filter = py7zr.compressor.BCJEncoder()
        m = hashlib.sha256()
        with tmp_path.joinpath("target.bin").open("wb") as fo:
            data = f.read(8192)
            while len(data) > 0:
                odata = filter.compress(data)
                m.update(odata)
                fo.write(odata)
                data = f.read(8192)
            odata = filter.flush()
            m.update(odata)
            fo.write(odata)
        assert m.digest() == binascii.unhexlify("40cb9389e9eb800540389321cace40866168a2bee563e404be680c8dda3f3d4f")


@pytest.mark.unit
def test_bcj_decode():
    with open(os.path.join(testdata_path, "bcj.bin"), "rb") as f:
        filter = py7zr.compressor.BCJDecoder(1052)
        m = hashlib.sha256()
        data = f.read(8192)
        while len(data) > 0:
            m.update(filter.decompress(data))
            data = f.read(8192)
        assert m.digest() == binascii.unhexlify("10c9fae2722a8dab791fd3a59393bd8ad00121db7088cd86fd8bcd36f4a12e65")


@pytest.mark.unit
def test_remove_relative_path_marker():
    filename = "relative_file"
    relative_file_path = f"{py7zr.helpers.RELATIVE_PATH_MARKER}{filename}"
    assert py7zr.helpers.remove_relative_path_marker(relative_file_path) == filename


@pytest.mark.unit
def test_remove_relative_path_marker_no_change():
    absolute_file_path = "/absolute/path/file"
    assert py7zr.helpers.remove_relative_path_marker(absolute_file_path) == absolute_file_path


@pytest.mark.unit
def test_remove_relative_path_marker_ignore_marker():
    file_path = f"/path/with/ignored/marker{py7zr.helpers.RELATIVE_PATH_MARKER}"
    # The marker isn't removed if it isn't at the beginning
    assert py7zr.helpers.remove_relative_path_marker(file_path) == file_path



import ctypes
import io
import os
import pathlib
import shutil
import subprocess
import sys
from contextlib import contextmanager

import multivolumefile
import pytest

import py7zr

from . import libarchive_extract, p7zip_test

testdata_path = pathlib.Path(os.path.dirname(__file__)).joinpath("data")
os.umask(0o022)


@pytest.mark.misc
def test_extract_multi_volume(tmp_path):
    with testdata_path.joinpath("lzma2bcj.7z").open("rb") as src:
        with tmp_path.joinpath("lzma2bcj.7z.001").open("wb") as tgt:
            tgt.write(src.read(25000))
        with tmp_path.joinpath("lzma2bcj.7z.002").open("wb") as tgt:
            tgt.write(src.read(27337))
    with multivolumefile.open(tmp_path.joinpath("lzma2bcj.7z"), mode="rb") as tgt:
        with py7zr.SevenZipFile(tgt) as arc:
            arc.extractall(tmp_path.joinpath("tgt"))


@pytest.mark.misc
def test_compress_to_multi_volume(tmp_path):
    tmp_path.joinpath("src").mkdir()
    py7zr.unpack_7zarchive(os.path.join(testdata_path, "lzma2bcj.7z"), path=tmp_path.joinpath("src"))
    with multivolumefile.open(tmp_path.joinpath("target.7z"), mode="wb", volume=10240) as tgt:
        with py7zr.SevenZipFile(tgt, "w") as arc:
            arc.writeall(tmp_path.joinpath("src"), "src")
    assert tmp_path.joinpath("target.7z.0001").stat().st_size == 10240
    assert tmp_path.joinpath("target.7z.0002").stat().st_size == 10240
    assert tmp_path.joinpath("target.7z.0003").stat().st_size == 10240
    assert 6000 < tmp_path.joinpath("target.7z.0004").stat().st_size < 6100
    #
    p7zip_test(tmp_path.joinpath("target.7z.0001"))


@pytest.mark.misc
def test_copy_bcj_file(tmp_path):
    with py7zr.SevenZipFile(testdata_path.joinpath("copy_bcj_1.7z").open(mode="rb")) as ar:
        ar.extractall(tmp_path)


@pytest.mark.misc
@pytest.mark.skipif(sys.platform.startswith("win"), reason="Case uses gcc and posix path on linux/mac")
def test_bcj_file(tmp_path):
    tmp_path.joinpath("src").mkdir()
    # build test target data
    if shutil.which("gcc"):
        result = subprocess.run(
            [
                "gcc",
                "-std=c99",
                "-fPIC",
                "-o",
                tmp_path.joinpath("src").joinpath("bcj_test").as_posix(),
                "-c",
                testdata_path.joinpath("bcj_test.c").as_posix(),
            ],
            stdout=subprocess.PIPE,
        )
        if result.returncode != 0:
            return 0
        #
        tmp_path.joinpath("tgt").mkdir()
        tmp_path.joinpath("tgt2").mkdir()
        my_filters = [{"id": py7zr.FILTER_X86}, {"id": py7zr.FILTER_COPY}]
        with py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), filters=my_filters, mode="w") as ar:
            ar.write(tmp_path.joinpath("src/bcj_test"), "bcj_test")
        target = tmp_path.joinpath("target.7z")
        with py7zr.SevenZipFile(target, "r") as ar:
            ar.extractall(tmp_path.joinpath("tgt"))
        p7zip_test(target)
        libarchive_extract(target, tmp_path / "tgt2")


@pytest.mark.files
@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
def test_double_extract_symlink(tmp_path):
    with py7zr.SevenZipFile(testdata_path.joinpath("symlink_2.7z").open(mode="rb")) as archive:
        archive.extractall(path=tmp_path)
    with py7zr.SevenZipFile(testdata_path.joinpath("symlink_2.7z").open(mode="rb")) as archive:
        archive.extractall(path=tmp_path)


class ProgressCallbackExample(py7zr.callbacks.ExtractCallback):
    def __init__(self):
        pass


def test_callback_raw_class():
    # test the case when passed argument is class name.
    # it is wrong, so it become the error.
    with pytest.raises(ValueError):
        with py7zr.SevenZipFile(testdata_path.joinpath("solid.7z").open(mode="rb")) as z:
            z.extractall(None, callback=ProgressCallbackExample)


def test_callback_not_concrete_class():
    # test the case when passed argument is abstract class
    with pytest.raises(TypeError):
        with py7zr.SevenZipFile(testdata_path.joinpath("solid.7z").open(mode="rb")) as z:
            cb = ProgressCallbackExample()
            z.extractall(None, callback=cb)


@pytest.mark.api
def test_extract_callback(tmp_path):
    # test the case when good callback passed.
    class ECB(py7zr.callbacks.ExtractCallback):
        def __init__(self, ofd):
            self.ofd = ofd

        def report_start_preparation(self):
            self.ofd.write("preparation.\n")

        def report_start(self, processing_file_path, processing_bytes):
            self.ofd.write('start "{}" (compressed in {} bytes)\n'.format(processing_file_path, processing_bytes))

        def report_update(self, decompressed_bytes):
            self.ofd.write("decompressed part of {} bytes\n".format(decompressed_bytes))

        def report_end(self, processing_file_path, wrote_bytes):
            self.ofd.write('end "{}" extracted to {} bytes\n'.format(processing_file_path, wrote_bytes))

        def report_postprocess(self):
            self.ofd.write("post processing.\n")

        def report_warning(self, message):
            self.ofd.write("warning: {:s}\n".format(message))

    cb = ECB(sys.stdout)
    with py7zr.SevenZipFile(open(os.path.join(testdata_path, "test_1.7z"), "rb")) as archive:
        archive.extractall(path=tmp_path, callback=cb)


@pytest.mark.misc
@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="Only meaningful only on Unix-like OSes",
)
@pytest.mark.slow
def test_extract_high_compression_rate(tmp_path):
    gen = Generator()
    with py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), "w") as source:
        source.writef(gen, "source")
    limit = int(512e6)  # 0.5GB
    with limit_memory(limit):
        with py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), "r") as target:
            target.extractall(path=tmp_path)


@contextmanager
def limit_memory(maxsize: int):
    """
    Decorator to limit memory. Raise MemoryError when the limit is exceeded.

    :param maxsize: Maximum size of memory resource to limit
    :raises: MemoryError: When function reaches the limit.
    """
    if sys.platform.startswith("win") or sys.platform.startswith("cygwin"):
        yield
    else:
        import resource

        import psutil  # type: ignore

        soft, hard = resource.getrlimit(resource.RLIMIT_AS)
        soft_new = psutil.Process(os.getpid()).memory_info().rss + maxsize
        if soft == -1 or soft_new < soft:
            resource.setrlimit(resource.RLIMIT_AS, (soft_new, hard))
        try:
            yield
        finally:
            resource.setrlimit(resource.RLIMIT_AS, (soft, hard))


class Generator(io.BufferedIOBase):
    def __init__(self):
        super(Generator, self).__init__()
        self.generated = 0
        self.length = int(1e9)
        self.seeked = False

    def readinto(self, b):
        if self.generated > self.length:
            return b""
        buf = self.generate_raw_bytes(len(b))
        b[: len(buf)] = buf
        self.generated += len(buf)
        return len(buf)

    def read1(self, size=-1):
        return self.read(size)

    def read(self, size=-1):
        if self.generated > self.length:
            return b""
        self.generated += size
        return self.generate_raw_bytes(size)

    def seek(self, *args, **kwargs):
        self.seeked = True

    def tell(self) -> int:
        if self.seeked:
            return self.length
        else:
            return 0

    def generate_raw_bytes(self, size):
        return bytes(size)



import argparse
import json
import pathlib
import sys

from tabulate import tabulate  # type: ignore


def generate_metainfo(root: dict) -> str:
    machine_info = root['machine_info']
    commit_info = root['commit_info']
    result = 'Machine: {} {} on {}({})\n'.format(machine_info['system'], machine_info['release'],
                                                 machine_info['cpu']['brand'],
                                                 machine_info['cpu']['hz_actual_friendly'])
    result += 'Python: {} {} [{} {}]\n'.format(machine_info["python_implementation"],
                                               machine_info["python_version"],
                                               machine_info["python_compiler"], machine_info["machine"])
    result += 'Commit: {} on {} in {}\n'.format(commit_info['id'], commit_info['branch'], commit_info['time'])
    return result


def generate_table(benchmarks: dict, group: str, type='simple') -> str:
    base = 1
    for bm in benchmarks:
        if group == bm['group']:
            if bm['params']['name'] == 'lzma2+bcj':
                base = bm['extra_info']['rate'] / 1000000
                if base < 0.1:
                    base = 0.1
    table = []
    for bm in benchmarks:
        if group == bm['group']:
            target = bm['params']['name']
            rate = bm['extra_info']['rate'] / 1000000
            if rate < 10:
                rate = round(rate, 2)
            else:
                rate = round(rate, 1)
            ratex = 'x {}'.format(round(rate / base, 2))
            ratio = round(float(bm['extra_info']['ratio']) * 100, 1)
            min = bm['stats']['min']
            max = bm['stats']['max']
            avr = bm['stats']['mean']
            table.append([target, rate, ratex, ratio, min, max, avr])
    return tabulate(table, headers=['target', 'speed(MB/sec)', 'rate', 'ratio(%)', 'min(sec)', 'max(sec)', 'mean(sec)'],
                    tablefmt=type)


def generate_comment(results_file, type):
    with open(results_file, 'r') as results:
        root = json.load(results)
    benchmarks = root['benchmarks']
    comment_body = '## Benchmark results\n\n'
    comment_body += generate_metainfo(root)
    comment_body += '\n\n### Compression benchmarks\n\n'
    comment_body += generate_table(benchmarks, 'compress', type=type)
    comment_body += '\n\n### Decompression benchmarks\n\n'
    comment_body += generate_table(benchmarks, 'decompress', type=type)
    return comment_body


def main():
    parser = argparse.ArgumentParser(prog='benchmark_result')
    parser.add_argument('jsonfile', type=pathlib.Path, help='pytest-benchmark saved result.')
    parser.add_argument('--markdown', action='store_true', help='print markdown table')
    args = parser.parse_args()
    if args.markdown:
        type = 'github'
    else:
        type = 'simple'
    body = generate_comment(args.jsonfile, type)
    print(body)


if __name__ == "__main__":
    sys.exit(main())



# Security protection test cases
import ctypes
import os
import pathlib
import sys
from dataclasses import dataclass
from tempfile import TemporaryDirectory

import pytest

from py7zr import SevenZipFile
from py7zr.exceptions import Bad7zFile
from py7zr.helpers import check_archive_path, get_sanitized_output_path
from py7zr.properties import FILTER_LZMA2, PRESET_DEFAULT

testdata_path = os.path.join(os.path.dirname(__file__), "data")


@pytest.mark.misc
def test_check_archive_path():
    bad_path = "../../.../../../../../../tmp/evil.sh"
    assert not check_archive_path(bad_path)


@pytest.mark.misc
def test_get_sanitized_output_path_1(tmp_path):
    bad_path = "../../.../../../../../../tmp/evil.sh"
    with pytest.raises(Bad7zFile):
        get_sanitized_output_path(bad_path, tmp_path)


@pytest.mark.misc
def test_get_sanitized_output_path_2(tmp_path):
    good_path = "good.sh"
    expected = tmp_path.joinpath(good_path)
    assert expected == get_sanitized_output_path(good_path, tmp_path)


@pytest.mark.misc
def test_extract_path_traversal_attack(tmp_path):
    my_filters = [
        {"id": FILTER_LZMA2, "preset": PRESET_DEFAULT},
    ]
    target = tmp_path.joinpath("target.7z")
    good_data = b"#!/bin/sh\necho good\n"
    good_path = "good.sh"
    bad_data = b"!#/bin/sh\necho bad\n"
    bad_path = "../../.../../../../../../tmp/evil.sh"
    with SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.writestr(good_data, good_path)
        archive._writestr(bad_data, bad_path)  # bypass a path check
    with pytest.raises(Bad7zFile):
        with SevenZipFile(target, "r") as archive:
            archive.extractall(path=tmp_path)


@pytest.mark.skipif(
    sys.platform.startswith("win") and (ctypes.windll.shell32.IsUserAnAdmin() == 0),
    reason="Administrator rights is required to make symlink on windows",
)
@pytest.mark.misc
def test_extract_symlink_attack(tmp_path):
    my_filters = [
        {"id": FILTER_LZMA2, "preset": PRESET_DEFAULT},
    ]
    source_dir = tmp_path / "src"
    symlink_file = source_dir / "symlink.sh"
    source_dir.mkdir(exist_ok=True)
    target_dir = tmp_path / "tgt"
    target = tmp_path / "target.7z"
    target_dir.mkdir(exist_ok=True)
    bad_data = b"!#/bin/sh\necho bad\n"
    bad_path = tmp_path.joinpath("evil.sh")
    with bad_path.open("wb") as evil:
        evil.write(bad_data)
    symlink_file.symlink_to(bad_path)
    with SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.writeall(source_dir, "src")
    with pytest.raises(Bad7zFile):
        with SevenZipFile(target, "r") as archive:
            archive.extractall(path=target_dir)


def test_write_compressed_archive(tmp_path):
    @dataclass
    class Contents:
        filename: str
        text: str

    contents = (Contents(filename="bin/qmake", text="qqqqq"), Contents(filename="lib/libhoge.so", text="hoge"))
    with TemporaryDirectory() as temp_path, SevenZipFile(
        tmp_path / "tools_qtcreator-linux-qt.tools.qtcreator.7z", "w"
    ) as archive:  # fmt: skip
        dest = pathlib.Path(temp_path)
        for folder in ("bin", "lib", "mkspecs"):
            (dest / folder).mkdir(parents=True, exist_ok=True)
        for f in contents:
            full_path = dest / f.filename
            if not full_path.parent.exists():
                full_path.parent.mkdir(parents=True)
            full_path.write_text(f.text, "utf_8")
        archive.writeall(path=temp_path, arcname="target")
    with TemporaryDirectory() as target_path, SevenZipFile(
        tmp_path / "tools_qtcreator-linux-qt.tools.qtcreator.7z", "r"
    ) as archive:  # fmt: skip
        archive.extractall(path=target_path)



import io
import os
import pathlib
from typing import Optional, Union

import pytest

import py7zr
import py7zr.io
from py7zr import SevenZipFile

from . import p7zip_test

testdata_path = pathlib.Path(os.path.dirname(__file__)).joinpath("data")
os.umask(0o022)


class TestArchiveWriter(py7zr.io.Py7zIO):

    def __init__(self, fname, target: SevenZipFile):
        self.fname = fname
        self.target = target
        self.buffer = io.BytesIO()

    def write(self, s: Union[bytes, bytearray]) -> int:
        return self.buffer.write(s)

    def read(self, size: Optional[int] = None) -> bytes:
        return self.buffer.read(size)

    def seek(self, offset: int, whence: int = 0) -> int:
        return self.buffer.seek(offset, whence)

    def flush(self) -> None:
        pass

    def close(self) -> None:
        self.target.writef(self.buffer, self.fname)

    def size(self) -> int:
        return self.buffer.getbuffer().nbytes


class TestWriterFactory(py7zr.io.WriterFactory):
    def __init__(self, target: SevenZipFile):
        self.target = target

    def create(self, filename: str) -> py7zr.io.Py7zIO:
        return TestArchiveWriter(filename, self.target)


@pytest.mark.files
def test_read_write_new(tmp_path):
    with py7zr.SevenZipFile(tmp_path.joinpath("target.7z"), "w") as target:
        with py7zr.SevenZipFile(testdata_path.joinpath("mblock_1.7z").open(mode="rb")) as source:
            source.extractall(factory=TestWriterFactory(target))
    p7zip_test(tmp_path / "target.7z")



#!/usr/bin/python -u
#
# p7zr library
#
# Copyright (c) 2020,2021 Hiroshi Miura <miurahr@linux.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from abc import ABC, abstractmethod


class Callback(ABC):
    """Abstract base class for progress callbacks."""

    @abstractmethod
    def report_start_preparation(self):
        """report a start of preparation event such as making list of files and looking into its properties."""
        pass  # noqa

    @abstractmethod
    def report_start(self, processing_file_path, processing_bytes):
        """report a start event of specified archive file and its input bytes."""
        pass  # noqa

    @abstractmethod
    def report_update(self, decompressed_bytes):
        """report an event when large file is being extracted more than 1 second or when extraction is finished.
        receives a number of decompressed bytes since the last update."""
        pass  # noqa

    @abstractmethod
    def report_end(self, processing_file_path, wrote_bytes):
        """report an end event of specified archive file and its output bytes."""
        pass  # noqa

    @abstractmethod
    def report_warning(self, message):
        """report an warning event with its message"""
        pass  # noqa

    @abstractmethod
    def report_postprocess(self):
        """report a start of post processing event such as set file properties and permissions or creating symlinks."""
        pass  # noqa


class ExtractCallback(Callback):
    """Abstract base class for extraction progress callbacks."""

    pass


class ArchiveCallback(Callback):
    """Abstract base class for progress callbacks."""

    pass



import os

import pytest

from py7zr import SevenZipFile

testdata_path = os.path.join(os.path.dirname(__file__), "data")


def test_getinfo_file():
    with SevenZipFile(os.path.join(testdata_path, "copy.7z")) as archive:
        member = archive.getinfo("test1.txt")
        assert member.filename == "test1.txt"
        assert member.is_directory is False


def test_getinfo_dir():
    with SevenZipFile(os.path.join(testdata_path, "copy.7z")) as archive:
        member = archive.getinfo("test")
        assert member.filename == "test"
        assert member.is_directory is True


def test_getinfo_file_with_trailing_slash():
    with SevenZipFile(os.path.join(testdata_path, "copy.7z")) as archive:
        member = archive.getinfo("test1.txt/")
        assert member.filename == "test1.txt"
        assert member.is_directory is False


def test_getinfo_dir_with_trailing_slash():
    with SevenZipFile(os.path.join(testdata_path, "copy.7z")) as archive:
        member = archive.getinfo("test/")
        assert member.filename == "test"
        assert member.is_directory is True


def test_getinfo_missing_member():
    with pytest.raises(KeyError):
        with SevenZipFile(os.path.join(testdata_path, "copy.7z")) as archive:
            archive.getinfo("doesn't exist")



import hashlib
import io
import os
import pathlib
import sys
import zipfile
import zlib

import pytest
from Cryptodome.Cipher import AES

import py7zr
import py7zr.compressor
from py7zr import UnsupportedCompressionMethodError
from py7zr.properties import FILTER_DEFLATE64
from tests import p7zip_test

testdata_path = pathlib.Path(os.path.dirname(__file__)).joinpath("data")
srcdata = testdata_path.joinpath("src.zip")
os.umask(0o022)


@pytest.mark.unit
def test_aes_cipher():
    key = b"e\x11\xf1Pz<*\x98*\xe6\xde\xf4\xf6X\x18\xedl\xf2Be\x1a\xca\x19\xd1\\\xeb\xc6\xa6z\xe2\x89\x1d"
    iv = b"|&\xae\x94do\x8a4\x00\x00\x00\x00\x00\x00\x00\x00"
    indata = (
        b"T\x9f^\xb5\xbf\xdc\x08/\xfe<\xe6i'\x84A^\x83\xdc\xdd5\xe9\xd5\xd0b\xa9\x7fH$\x11\x82\x8d"
        b"\xce[\x85\xe7\xf2}\xe3oJ*\xc0:\xf4\xfd\x82\xe8I"
    )
    expected = (
        b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
        b"\xf7\\m\xe0t\xa6$yF_-\xa0\x0b8f "
    )
    cipher = AES.new(key, AES.MODE_CBC, iv)
    result = cipher.decrypt(indata)
    assert expected == result


@pytest.mark.unit
def test_deflate_compressor():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    compressor = py7zr.compressor.DeflateCompressor()
    outdata = compressor.compress(plain_data)
    outdata += compressor.flush()
    expected = b"c\xd0\x92\xe2TO\x91\xdc`Q|\xaa[Xa\xbd\x95t\xef\xf4\x1f5\xca\xbe/\x1f\x86_y2\xdd\xc9\xe9\x12\x03\x01y\x00"
    assert outdata == expected


@pytest.mark.unit
def test_deflate_compressor_flushed():
    with pytest.raises(zlib.error):
        plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
        plain_data += plain_data
        compressor = py7zr.compressor.DeflateCompressor()
        outdata = compressor.compress(plain_data)
        outdata += compressor.flush()
        outdata += compressor.flush()


@pytest.mark.unit
def test_deflate_decompressor():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    source_data = b"c\xd0\x92\xe2TO\x91\xdc`Q|\xaa[Xa\xbd\x95t\xef\xf4\x1f5\xca\xbe/\x1f\x86_y2\xdd\xc9\xe9\x12\x03\x01y\x00"
    decompressor = py7zr.compressor.DeflateDecompressor()
    outdata = decompressor.decompress(source_data)
    assert outdata == plain_data


@pytest.mark.unit
def test_copy_compressor():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    compressor = py7zr.compressor.CopyCompressor()
    outdata = compressor.compress(plain_data)
    outdata += compressor.flush()
    assert outdata == plain_data


@pytest.mark.unit
def test_copy_decompressor():
    source_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    decompressor = py7zr.compressor.CopyDecompressor()
    outdata = decompressor.decompress(source_data)
    assert outdata == source_data


@pytest.mark.unit
def test_pyzstd_decompressor_1():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    compressed = (
        b"\x28\xb5\x2f\xfd"
        b"\x00XE\x01\x00\x04\x02\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4"
        b"\xe4\x97BB\xd2\x01\x00\x18\xb8z\x02"
    )
    property = b"\x01\x04\x04\x00\x00"
    decompressor = py7zr.compressor.ZstdDecompressor(property, None)
    outdata = decompressor.decompress(compressed)
    assert outdata == plain_data


@pytest.mark.unit
def test_pyzstd_decompressor_2():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    compressed_0 = b"\x28\xb5\x2f\xfd"
    compressed_1 = b" @E\x01\x00\x04\x02\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4"
    compressed_2 = b"\xe4\x97BB\xd2\x01\x00\x18\xb8z\x02"
    property = b"\x01\x04\x04\x00\x00"
    decompressor = py7zr.compressor.ZstdDecompressor(property, None)
    outdata = decompressor.decompress(compressed_0)
    outdata += decompressor.decompress(compressed_1)
    outdata += decompressor.decompress(compressed_2)
    assert outdata == plain_data


@pytest.mark.unit
def test_sevenzipcompressor_aes_lzma2():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data + plain_data
    filters = [
        {"id": py7zr.FILTER_LZMA2, "preset": py7zr.PRESET_DEFAULT},
        {"id": py7zr.FILTER_CRYPTO_AES256_SHA256},
    ]
    indata = io.BytesIO(plain_data)
    outdata = io.BytesIO()
    compressor = py7zr.compressor.SevenZipCompressor(filters=filters, password="secret")
    insize, outsize, crc = compressor.compress(indata, outdata)
    outsize += compressor.flush(outdata)
    outdata.seek(0, 0)
    coders = compressor.coders
    unpacksizes = compressor.unpacksizes
    decompressor = py7zr.compressor.SevenZipDecompressor(
        coders=coders,
        packsize=outsize,
        unpacksizes=unpacksizes,
        crc=None,
        password="secret",
    )
    revert_data = decompressor.decompress(outdata)
    assert revert_data == plain_data


@pytest.mark.unit
def test_aes_compressor():
    compressor = py7zr.compressor.AESCompressor("secret")
    assert compressor.method == py7zr.properties.CompressionMethod.CRYPT_AES256_SHA256
    assert len(compressor.encode_filter_properties()) == 2 + 16


@pytest.mark.unit
def test_sevenzipcompressor_aes_only():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
    plain_data += plain_data
    filters = [{"id": py7zr.FILTER_CRYPTO_AES256_SHA256}]
    compressor = py7zr.compressor.SevenZipCompressor(filters=filters, password="secret")
    indata = io.BytesIO(plain_data)
    outdata = io.BytesIO()
    _, _, _ = compressor.compress(indata, outdata)
    _ = compressor.flush(outdata)
    result = outdata.getvalue()
    assert len(result) == 64
    assert result != plain_data


@pytest.mark.unit
def test_aes_encrypt_data():
    plain_data = b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB"
    plain_data += plain_data + plain_data
    compressor = py7zr.compressor.AESCompressor("secret")
    outdata = compressor.compress(plain_data)
    outdata += compressor.flush()
    assert len(outdata) == 96  # 96 = 16 * 6 = len(plain_data) + 3


@pytest.mark.unit
def test_aes_decrypt(monkeypatch):
    properties = b"S\x07|&\xae\x94do\x8a4"
    indata = (
        b"T\x9f^\xb5\xbf\xdc\x08/\xfe<\xe6i'\x84A^\x83\xdc\xdd5\xe9\xd5\xd0b\xa9\x7fH$\x11\x82\x8d"
        b"\xce[\x85\xe7\xf2}\xe3oJ*\xc0:\xf4\xfd\x82\xe8I"
    )
    expected = (
        b"\x00*\x1a\t'd\x19\xb08s\xca\x8b\x13 \xaf:\x1b\x8d\x97\xf8|#M\xe9\xe1W\xd4\xe4\x97BB\xd2"
        b"\xf7\\m\xe0t\xa6$yF_-\xa0\x0b8f "
    )
    decompressor = py7zr.compressor.AESDecompressor(properties, "secret")
    assert decompressor.decompress(indata) == expected


@pytest.mark.files
def test_extract_bzip2(tmp_path):
    archive = py7zr.SevenZipFile(testdata_path.joinpath("bzip2_2.7z").open(mode="rb"))
    archive.extractall(path=tmp_path)
    archive.close()


@pytest.mark.files
def test_extract_ppmd(tmp_path):
    archive = py7zr.SevenZipFile(testdata_path.joinpath("ppmd.7z").open(mode="rb"))
    archive.extractall(path=tmp_path)
    archive.close()


@pytest.mark.files
def test_extract_ppmd2(tmp_path):
    with py7zr.SevenZipFile(testdata_path.joinpath("testdata-x5-ppmd.7z").open(mode="rb")) as archive:
        archive.extractall(path=tmp_path)


@pytest.mark.files
def test_extract_deflate(tmp_path):
    with py7zr.SevenZipFile(testdata_path.joinpath("deflate.7z").open(mode="rb")) as archive:
        archive.extractall(path=tmp_path)


@pytest.mark.basic
def test_compress_ppmd(tmp_path):
    my_filters = [{"id": py7zr.FILTER_PPMD, "order": 6, "mem": 24}]  # mem in 2^size
    target = tmp_path.joinpath("target.7z")
    archive = py7zr.SevenZipFile(target, "w", filters=my_filters)
    archive.writeall(os.path.join(testdata_path, "src"), "src")
    archive.close()
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_ppmd_2(tmp_path):
    with py7zr.SevenZipFile(testdata_path.joinpath("bzip2_2.7z").open(mode="rb")) as arc:
        arc.extractall(path=tmp_path)
    my_filters = [{"id": py7zr.FILTER_PPMD, "order": 6, "mem": "4m"}]
    target = tmp_path.joinpath("target.7z")
    with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.encoded_header_mode = False
        archive.writeall(tmp_path.joinpath("10000SalesRecords.csv"))
    #
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
def test_compress_decompress_brotli(tmp_path):
    my_filters = [{"id": py7zr.FILTER_BROTLI, "level": 11}]
    #
    with py7zr.SevenZipFile(testdata_path.joinpath("bzip2_2.7z").open(mode="rb")) as arc:
        arc.extractall(path=tmp_path)
    target = tmp_path.joinpath("target.7z")
    with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.write(tmp_path.joinpath("10000SalesRecords.csv"), "10000SalesRecords.csv")
    tmp_path.joinpath("tgt").mkdir(exist_ok=True)
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path.joinpath("tgt"))


@pytest.mark.basic
def test_decompress_brotli(tmp_path):
    with pytest.raises(UnsupportedCompressionMethodError):
        with py7zr.SevenZipFile(testdata_path.joinpath("zstdmt-brotli.7z").open(mode="rb")) as archive:
            archive.extractall(path=tmp_path.joinpath("tgt"))


@pytest.mark.basic
@pytest.mark.skipif(hasattr(sys, "pypy_version_info"), reason="No support for deflate64 on pypy")
def test_compress_deflate64(tmp_path):
    # prepare source data
    with py7zr.SevenZipFile(testdata_path.joinpath("bzip2_2.7z").open(mode="rb")) as arc:
        arc.extractall(path=tmp_path)
    # compress with deflate64
    target = tmp_path.joinpath("target.7z")
    my_filters = [{"id": FILTER_DEFLATE64}]
    with py7zr.SevenZipFile(target, "w", filters=my_filters) as archive:
        archive.write(tmp_path.joinpath("10000SalesRecords.csv"), "10000SalesRecords.csv")
    # check extract
    tmp_path.joinpath("tgt").mkdir(exist_ok=True)
    with py7zr.SevenZipFile(target, "r") as archive:
        archive.extractall(path=tmp_path.joinpath("tgt"))
    # check with p7zip
    p7zip_test(tmp_path / "target.7z")


@pytest.mark.basic
@pytest.mark.skipif(hasattr(sys, "pypy_version_info"), reason="No support for deflate64 on pypy")
def test_decompress_deflate64(tmp_path):
    with zipfile.ZipFile(srcdata) as srczip:
        srczip.extractall(path=tmp_path.joinpath("src"))
    with py7zr.SevenZipFile(testdata_path.joinpath("deflate64.7z").open(mode="rb")) as archive:
        archive.extractall(path=tmp_path.joinpath("tgt"))
    for target in tmp_path.joinpath("tgt").glob("test-file.*"):
        m0 = hashlib.sha256()
        m0.update(tmp_path.joinpath("tgt").joinpath(target).open("rb").read())
        m1 = hashlib.sha256()
        m1.update(tmp_path.joinpath("src").joinpath(target).open("rb").read())
        assert m0.digest() == m1.digest(), "Fails digest for %s" % target
